#include <cassert>
#include <iostream>
#include <Eigen/Dense>
#include <nag.h>
#include <dco.hpp>
#include <nagad.h>
#include <stdio.h>
#include <math.h>
#include <nagx02.h>
#include <iostream>
using namespace std;


struct data {
  Integer m;
  Integer n;
  double *X;
  double *y;
};

struct data parameters;

template <typename T>
T f(const Integer &n, const T *p, const T *x){
    T res = 0.0;

    for (Integer i=0; i<n; ++i){
        res +=p[i]*x[i];
    }
    res = pow(res,2);
    return res;
}


template <typename T>
void E(const T *p, const struct data *par, T *fvec){
    Integer m = par->m;
    Integer n = par->n;
    T tmp;
    for (Integer i = 0; i < m; i++) {
        tmp = f(n, p, &par->X[i*n]);
        fvec[i] = tmp-par->y[i];
    }
}


template <typename T>
void lsqfun(Integer &,
            const Integer &,
            const Integer &,
            const T xc[],
            T fvec[],
            Integer *,
            const Integer &,
            T *,
            const Integer &)
{
    E(xc, &parameters, fvec);
    return;
}


template <typename T>
void lsqmon(const Integer &m,
            const Integer &n,
            const T xc[],
            const T fvec[],
            const T *,
            const Integer &,
            const T *,
            const Integer &,
            const Integer &niter,
            const Integer &,
            Integer *,
            const Integer &,
            T *,
            const Integer &)
{
    if (niter==0)
        printf("Iteration   Function Value          Current  Solution\n");

    double fsumsq = 0.0;
    for (int i=0;i<m; i++) {
        double f = fvec[i];
        fsumsq = fsumsq + f*f;
    }

    printf ("%5d       %10.5f", niter, fsumsq);
    size_t max_out = n;
    if (max_out>5) max_out=5;

    for (size_t i=0; i<max_out; i++)
        printf ("          %.5f", xc[i]);

    printf("\n");

}


int main(int argc, char* argv[])
{
    assert(argc==3); Integer m=std::stoi(argv[1]), n=std::stoi(argv[2]);

    Eigen::Matrix<double, Eigen::Dynamic, 1> y;
    Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic> t;
    /* this is the data to be fitted */
    t = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Random(m,n);
    y = Eigen::Matrix<double, Eigen::Dynamic, 1>::Random(m);
    t=t.cwiseProduct(t);

    parameters.m = m; parameters.n = n;
    parameters.y = new double[m];
    for (Integer i=0; i<m; i++) parameters.y[i] = y(i);

    parameters.X = new double[m*n];
    for (Integer i=0; i<m; i++){
        for (Integer j=0; j<n; j++)
            parameters.X[i*n+j] = t(i,j);
    }

    // Scalars
    int               exit_status = 0;
    const Integer     ldfjac = m, ldv = n;
    Integer ifail = 0;

    Integer           iprint, maxcal, niter, nf;
    Integer liw=1, lw;
    lw = 6*n + m*n + 2*m + n*n;
    Integer *iw = new Integer[liw];
    double *w = new double[lw];
    double eta, stepmx, xtol, fsumsq;
    double *x = new double[n];
    double *fvec = new double[m];
    double *fjac = new double[m*n];

    double *s = new double[n];
    double *v = new double[ldv*n];




    iprint = 1;
    maxcal = 400*n;
    eta = 0.5;
    xtol = 10.0*sqrt(X02AJC);
    stepmx = 10.0;

    // Starting points
    for (Integer i=0; i<n; i++) x[i] = -1.0;

    ifail = -1;

    e04fcf_(m,n,lsqfun,lsqmon,iprint,maxcal,eta,xtol,
            stepmx,x,fsumsq,fvec,fjac,ldfjac,s,v,ldv,niter,nf,
            iw,liw,w,lw,ifail);

    cout << "\n  summary from method e04fc" << endl;
    cout << "  number of iterations: " << niter << endl;
    cout << "  function evaluations: " << nf << endl;
    // Primal results
    cout.setf(ios::fixed,ios::floatfield);
    if (ifail==0 || ifail>1) {
        cout.precision(6);
        cout << "\n  final |f(x)|     = ";
        cout.width(20); cout << fsumsq << endl;

        int max_out = n;
        if (max_out>5) max_out=5;
        cout << "  Minimum found at = ";
        for (int i=0; i<max_out; i++) {
            cout.width(20);
            cout << x[i];

        }
        cout << endl;

    }


    cout << "\n Derivatives calculated: Centered finite differences\n";
    cout << " Derivatives:\n\n";

    double fsumsq_ph, h, fsumsq_mh;

    iprint = -1;
    // Get derivatives of fsumsq
    cout << "  sum of squares w.r.t y:\n";
    for (Integer i=0; i<m; ++i){
        h=fabs(parameters.y[i])<1 ? sqrt(std::numeric_limits<double>::epsilon())
                       : sqrt(std::numeric_limits<double>::epsilon())*fabs(parameters.y[i]);
        // Starting points
        for (Integer j=0; j<n; j++) x[j] = -1.0;
        double tmp = parameters.y[i];
        parameters.y[i] += h;
        ifail = 1;
        e04fcf_(m,n,lsqfun,lsqmon,iprint,maxcal,eta,xtol,
                stepmx,x,fsumsq_ph,fvec,fjac,ldfjac,s,v,ldv,niter,nf,
                iw,liw,w,lw,ifail);

        // Starting points
        for (Integer j=0; j<n; j++) x[j] = -1.0;
        parameters.y[i] = tmp - h;
        ifail = 1;
        e04fcf_(m,n,lsqfun,lsqmon,iprint,maxcal,eta,xtol,
                stepmx,x,fsumsq_mh,fvec,fjac,ldfjac,s,v,ldv,niter,nf,
                iw,liw,w,lw,ifail);


        cout << "  dfsumsq/dy[" << i <<"]= " <<(fsumsq_ph-fsumsq_mh)/(2*h) << endl;
        parameters.y[i] = tmp;
    }
    cout << endl;

    delete [] parameters.X;
    delete [] parameters.y;
    delete [] iw;
    delete [] w;
    delete [] x;
    delete [] fvec;
    delete [] fjac;
    delete [] s;
    delete [] v;

    return exit_status;
}


